import openmdao.api as om
import dymos as dm

import unittest
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from dymos.utils.misc import om_version


class ODEComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        # z is the state vector, a nn x 2 x 2 in the form of [[x, y], [vx, vy]]
        self.add_input('z', shape=(nn, 2, 2), units=None)
        self.add_output('zdot', shape=(nn, 2, 2), units=None)

        self.declare_partials(of='zdot', wrt='z', method='cs')
        self.declare_coloring(wrt=['z'], method='cs', num_full_jacs=5, tol=1.0E-12)

    def compute(self, inputs, outputs):
        outputs['zdot'][:, 0, 0] = inputs['z'][:, 1, 0]
        outputs['zdot'][:, 0, 1] = inputs['z'][:, 1, 1]
        outputs['zdot'][:, 1, 0] = 0.0
        outputs['zdot'][:, 1, 1] = -9.81


@use_tempdirs
class TestCannonballMatrixState(unittest.TestCase):
    """ Tests to verify that dymos can use matrix-states"""

    def _make_problem(self, tx):
        p = om.Problem()

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ODEComp,
                         transcription=tx)
        traj.add_phase('phase', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(1, 5), units=None)
        phase.add_state('z', rate_source='zdot', fix_initial=True, units=None)

        phase.add_boundary_constraint('z', loc='final', equals=0, indices=[1])
        phase.add_path_constraint('z', lower=0, upper=1E3, indices=[0])
        phase.add_objective('time', loc='final')

        phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

        p.driver = om.pyOptSparseDriver(optimizer='IPOPT')
        p.driver.declare_coloring(tol=1.0E-12)

        p.setup()

        phase.set_time_val(initial=0, duration=5)
        phase.set_state_val('z', [[[0, 0], [10, 10]], [[10, 0], [10, -10]]])

        return p

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_radau(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        print(p.get_val('traj.phase.timeseries.z'))

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_gl(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_birkhoff_lgl(self):
        tx = dm.Birkhoff(num_nodes=20, grid_type='lgl')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_birkhoff_cgl(self):
        tx = dm.Birkhoff(num_nodes=20, grid_type='cgl')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_radau_solve_segments(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_cannonball_matrix_state_gl_solve_segments(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)


@use_tempdirs
class TestCannonballMatrixStateExplicitShape(unittest.TestCase):
    """ Tests to verify that dymos can use matrix-states"""

    def _make_problem(self, tx):
        p = om.Problem()

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ODEComp,
                         transcription=tx)
        traj.add_phase('phase', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(1, 5), units=None)
        phase.add_state('z', rate_source='zdot', fix_initial=True, units=None, shape=(2, 2))

        phase.add_boundary_constraint('z', loc='final', equals=0, indices=[1])
        phase.add_path_constraint('z', lower=0, upper=1E3, indices=[0])
        phase.add_objective('time', loc='final')

        phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        p.setup()

        phase.set_time_val(initial=0, duration=5)
        phase.set_state_val('z', [[[0, 0], [10, 10]], [[10, 0], [10, -10]]])

        return p

    def test_cannonball_matrix_state_radau(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_gl(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_radau_solve_segments(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_gl_solve_segments(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)


@use_tempdirs
class TestCannonballMatrixStateNonFlatIndices(unittest.TestCase):
    """ Tests to verify that dymos can use matrix-states"""

    def _make_problem(self, tx):
        p = om.Problem()

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ODEComp,
                         transcription=tx)
        traj.add_phase('phase', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(1, 5), units=None)
        phase.add_state('z', rate_source='zdot', fix_initial=True, units=None, shape=(2, 2))

        # When specifying a boundary constraint index using the coordinates within the matrix, use a slice.
        phase.add_boundary_constraint('z', loc='final', equals=0, indices=om.slicer[0, 1], flat_indices=False)
        phase.add_path_constraint('z', lower=0, upper=1E3, indices=om.slicer[0, 0], flat_indices=False)
        phase.add_objective('time', loc='final')

        phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        p.setup()

        phase.set_time_val(initial=0, duration=5)
        phase.set_state_val('z', [[[0, 0], [10, 10]], [[10, 0], [10, -10]]])

        return p

    def test_cannonball_matrix_state_radau(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_gl(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments=False)

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_radau_solve_segments(self):

        tx = dm.Radau(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

    def test_cannonball_matrix_state_gl_solve_segments(self):
        tx = dm.GaussLobatto(num_segments=10, order=3, solve_segments='forward')

        p = self._make_problem(tx)

        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(p.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        c = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(c.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 1], 0.0, tolerance=1E-5)
        assert_near_equal(c.get_val('traj.phase.timeseries.z')[-1, 0, 0], 20.3873598, tolerance=1E-5)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


def initial_guess(t_dur, gam0=np.pi/3):
    t = np.linspace(0, t_dur, num=100)
    g = -9.80665
    v0 = -0.5*g*t_dur/np.sin(gam0)
    r = v0*np.cos(gam0)*t
    h = v0*np.sin(gam0)*t + 0.5*g*t**2
    v = np.sqrt(v0*np.cos(gam0)**2 + (v0*np.sin(gam0) + g*t)**2)
    gam = np.arctan2(v0*np.sin(gam0) + g*t, v0*np.cos(gam0)**2)

    guess = {'t': t, 'r': r, 'h': h, 'v': v, 'gam': gam}

    return guess


@use_tempdirs
class TestCannonballImplicitDuration(unittest.TestCase):

    def test_cannonball_implicit_duration(self):
        import openmdao.api as om
        from scipy.interpolate import make_interp_spline
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.models.atmosphere.atmos_1976 import USatm1976Data

        #############################################
        # Component for the design part of the model
        #############################################
        class CannonballSizeComp(om.ExplicitComponent):
            """
            Compute the area and mass of a cannonball with a given radius and density.

            Notes
            -----
            This component is not vectorized with 'num_nodes' as is the usual way
            with Dymos, but is instead intended to compute a scalar mass and reference
            area from scalar radius and density inputs. This component does not reside
            in the ODE but instead its outputs are connected to the trajectory via
            input design parameters.
            """
            def setup(self):
                self.add_input(name='radius', val=1.0, desc='cannonball radius', units='m')
                self.add_input(name='dens', val=7870., desc='cannonball density', units='kg/m**3')

                self.add_output(name='mass', shape=(1,), desc='cannonball mass', units='kg')
                self.add_output(name='S', shape=(1,), desc='aerodynamic reference area', units='m**2')

                self.declare_partials(of='mass', wrt='dens')
                self.declare_partials(of='mass', wrt='radius')

                self.declare_partials(of='S', wrt='radius')

            def compute(self, inputs, outputs):
                radius = inputs['radius']
                dens = inputs['dens']

                outputs['mass'] = (4/3.) * dens * np.pi * radius ** 3
                outputs['S'] = np.pi * radius ** 2

            def compute_partials(self, inputs, partials):
                radius = inputs['radius']
                dens = inputs['dens']

                partials['mass', 'dens'] = (4/3.) * np.pi * radius ** 3
                partials['mass', 'radius'] = 4. * dens * np.pi * radius ** 2

                partials['S', 'radius'] = 2 * np.pi * radius

        #############################################
        # Build the ODE class
        #############################################
        class CannonballODE(om.ExplicitComponent):
            """
            Cannonball ODE assuming flat earth and accounting for air resistance
            """

            def initialize(self):
                self.options.declare('num_nodes', types=int)

            def setup(self):
                nn = self.options['num_nodes']

                # static parameters
                self.add_input('m', units='kg')
                self.add_input('S', units='m**2')
                # 0.5 good assumption for a sphere
                self.add_input('CD', 0.5)

                # time varying inputs
                self.add_input('h', units='m', shape=nn)
                self.add_input('v', units='m/s', shape=nn)
                self.add_input('gam', units='rad', shape=nn)

                # state rates
                self.add_output('v_dot', shape=nn, units='m/s**2', tags=['dymos.state_rate_source:v'])
                self.add_output('gam_dot', shape=nn, units='rad/s', tags=['dymos.state_rate_source:gam'])
                self.add_output('h_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:h'])
                self.add_output('r_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:r'])
                self.add_output('ke', shape=nn, units='J')

                # Ask OpenMDAO to compute the partial derivatives using finite-difference
                # with a partial coloring algorithm for improved performance, and use
                # a graph coloring algorithm to automatically detect the sparsity pattern.
                self.declare_coloring(wrt='*', method='fd')

                alt_data = USatm1976Data.alt * om.unit_conversion('ft', 'm')[0]
                rho_data = USatm1976Data.rho * om.unit_conversion('slug/ft**3', 'kg/m**3')[0]
                self.rho_interp = make_interp_spline(alt_data, rho_data, k=1)

            def compute(self, inputs, outputs):

                gam = inputs['gam']
                v = inputs['v']
                h = inputs['h']
                m = inputs['m']
                S = inputs['S']
                CD = inputs['CD']

                GRAVITY = 9.80665  # m/s**2

                rho = self.rho_interp(h)

                q = 0.5*rho*v**2
                qS = q * S
                D = qS * CD
                cgam = np.cos(gam)
                sgam = np.sin(gam)
                outputs['v_dot'] = - D/m-GRAVITY*sgam
                outputs['gam_dot'] = -(GRAVITY/v)*cgam
                outputs['h_dot'] = v*sgam
                outputs['r_dot'] = v*cgam
                outputs['ke'] = 0.5*m*v**2

        #############################################
        # Setup the Dymos problem
        #############################################

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(),
                              promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10,
                               ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=20, order=3, compressed=False)
        phase = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        phase = traj.add_phase('phase', phase)

        # All initial states except flight path angle are fixed
        # The output of the ODE which provides the rate source for each state
        # is obtained from the tags used on those outputs in the ODE.
        # The units of the states are automatically inferred by multiplying the units
        # of those rates by the time units.
        phase.set_time_options(fix_initial=True, duration_bounds=(1, 100), units='s')
        phase.add_state('r', fix_initial=True, solve_segments='forward')
        phase.add_state('h', fix_initial=True, solve_segments='forward')
        phase.add_state('gam', fix_initial=False, initial_bounds=(0, np.pi/2), solve_segments='forward')
        phase.add_state('v', fix_initial=False, solve_segments='forward')

        phase.add_parameter('S', units='m**2', static_target=True)
        phase.add_parameter('m', units='kg', static_target=True)
        phase.add_parameter('CD', val=0.5, opt=False, static_target=True)

        phase.add_boundary_constraint('ke', loc='initial',
                                      upper=400000, lower=0, ref=100000)

        # A duration balance is added setting altitude to zero.
        # A nonlinear solver is used to find the duration of required to satisfy the condition.
        # The duration was bounded to be greater than 1 to ensure the solver did not
        # converge to the initial point.
        phase.set_duration_balance('h', val=0.0)

        # In this problem, the default ArmijoGoldsteinLS has issues with extrapolating
        # the states and causes the optimization to fail.
        # Using the default linesearch or BoundsEnforceLS work better here.
        phase.nonlinear_solver = om.NewtonSolver(iprint=0, solve_subsystems=True,
                                                 maxiter=100, stall_limit=3)
        phase.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        phase.linear_solver = om.DirectSolver()

        phase.add_objective('r', loc='final', scaler=-1.0)

        p.model.connect('size_comp.mass', 'traj.phase.parameters:m')
        p.model.connect('size_comp.S', 'traj.phase.parameters:S')

        # Finish Problem Setup
        p.setup()

        #############################################
        # Set constants and initial guesses
        #############################################
        p.set_val('radius', 0.05, units='m')
        p.set_val('dens', 7.87, units='g/cm**3')

        p.set_val('traj.phase.parameters:CD', 0.5)

        guess = initial_guess(t_dur=30.0)

        phase.set_time_val(initial=0.0, duration=30.0)

        phase.set_state_val('r', vals=guess['r'], time_vals=guess['t'])
        phase.set_state_val('h', vals=guess['h'], time_vals=guess['t'])
        phase.set_state_val('v', vals=guess['v'], time_vals=guess['t'])
        phase.set_state_val('gam', vals=guess['gam'], time_vals=guess['t'], units='rad')

        #####################################################
        # Run the optimization and final explicit simulation
        #####################################################
        dm.run_problem(p, simulate=True)

        assert_near_equal(p.get_val('traj.phase.states:r')[-1],
                          3183.25, tolerance=1.0)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



"""
Utilities for dealing with ODE systems.
"""
from copy import deepcopy
import re

import numpy as np

import openmdao.api as om

from dymos.utils.misc import _unspecified


# This regex finds variables and any indices that follow them.
# 1 - leading underscore or letter (one)
# 2 - alphanumeric, underscore, period, or colon (zero or more)
# 3 - end of word is not followed by opening parentheses
# 4 - optional index specification in brackets
#                       1111111112222222333333334444444444444444
var_rgx = re.compile(r'([_a-zA-Z][\w.:]*\b(?!\()(\[[\d:.,-]+\])*)')


def _parse_index_string(index_str):
    """
    Convert a string representation of array indices to a tuple of slice objects,
    integer indices, or ellipsis that can be used for array access.

    Args:
        index_str (str): String representation of indices, e.g. '[1, 3]', '[1:5, 2]', '[..., 0]'

    Returns:
        tuple: A tuple containing slice objects, integers, or Ellipsis for indexing
    """
    if index_str is None:
        return None
    # Remove the outer brackets and whitespace
    clean_str = index_str.strip().strip('[]')

    if not clean_str:
        return tuple()

    # Split by comma to get each dimension's index
    dimensions = clean_str.split(',')

    result = []
    for dim in dimensions:
        dim = dim.strip()

        # Handle ellipsis
        if dim == '...' or dim == 'Ellipsis':
            result.append(Ellipsis)
            continue

        # Handle full slice ':'
        if dim == ':':
            result.append(slice(None))
            continue

        # Check if it's a slice (contains ':')
        if ':' in dim:
            # Handle slice notation (start:stop:step)
            slice_parts = dim.split(':')

            # Convert each part, handling empty strings
            start = int(slice_parts[0]) if slice_parts[0].strip() else None

            if len(slice_parts) > 1:
                stop = int(slice_parts[1]) if slice_parts[1].strip() else None
            else:
                stop = None

            if len(slice_parts) > 2:
                step = int(slice_parts[2]) if slice_parts[2].strip() else None
            else:
                step = None

            result.append(slice(start, stop, step))
        else:
            # It's a simple index
            try:
                result.append(int(dim))
            except ValueError:
                # If it's not an integer, it might be a variable
                result.append(dim)

    return tuple(result)


class ExprParser():
    """
    A special parser for Expressions in dymos.

    When invoked with a full_path as given by _extract_openmdao_variables,
    return a legal name for the variable in the expression and an indexer
    if indices were specified.
    """
    def __init__(self):
        self._unique_int = 0

    def parse(self, full_path, idx_str):
        """
        Extract the legal name and any indexer given by the full path representation.

        Parameters
        ----------
        full_path : str
            The full path of the expression variable. This may include
            dots, colons, and an index specification on the end.
        idx_str : str
            The string which follows the path specification in the expression.
            This begins and ends with brackets [].

        Returns
        -------
        tuple(str, tuple or None)
            The string containing the OpenMDAO-legal name for the variable
            in the ExecComp expression, as well as any indices specified.
        """
        is_ode_rel = False
        if '.' in full_path:
            # the path is ODE relative
            last_path = full_path.split('.')[-1]
            is_ode_rel = True
        else:
            # its a top-level variable or phase variable
            last_path = full_path
        legal_name = last_path.split(':')[-1]
        if is_ode_rel:
            legal_name += str(self._unique_int)
            self._unique_int += 1
        return legal_name, _parse_index_string(idx_str)


class ODEGroup(om.Group):
    """
    Initialize an ODEGroup to wrap the user's ODE and an ExecComp.

    Parameters
    ----------
    ode_class : class
        The class of ODE used in this Group.
    num_nodes : int
        The number of nodes used in the ODE.
    ode_init_kwargs : dict or None
        Initialization arguments for the ODE system.
    calc_exprs : dict
        The _calc_exprs dictionary of the owning phase instance.
    parameter_options : dict
        The parameter_options dictionary of the owning phase instance.
    """
    def __init__(self, ode_class, num_nodes, ode_init_kwargs=None, calc_exprs=None, parameter_options=None):
        super().__init__()
        self._ode_class = ode_class
        self._ode_init_kwargs = ode_init_kwargs or {}
        self._calc_exprs = calc_exprs or {}
        self._num_nodes = num_nodes
        self._parameter_options = parameter_options

    def setup(self):
        """
        Set up the ODEGroup.

        Adds the user's ODE class and an ExecComp.
        """
        ode_class = self._ode_class
        ode_init_kwargs = self._ode_init_kwargs
        num_nodes = self._num_nodes

        ode = ode_class(num_nodes=num_nodes, **ode_init_kwargs)

        self.add_subsystem('user_ode', ode, promotes_inputs=['*'], promotes_outputs=['*'])
        ec = om.ExecComp()
        self.add_subsystem('exec_comp', ec, promotes_inputs=['*'], promotes_outputs=['*'])

    def configure(self):
        """
        Setup up connections and promotions in the ODEGroup.
        """
        num_nodes = self._num_nodes

        seen_kwargs = set()
        parser = ExprParser()
        ec = self._get_subsystem('exec_comp')

        for expr, expr_kwargs in self._calc_exprs.items():
            common_units = _unspecified
            common_shape = _unspecified
            if 'units' in expr_kwargs:
                # units are set throughout this expression
                common_units = expr_kwargs['units']
            if 'shape' in expr_kwargs:
                common_shape = expr_kwargs['shape']

            _expr_kwargs = {}
            output_var, rhs = [s.strip() for s in expr.split('=')]
            _expr_kwargs[output_var] = deepcopy(expr_kwargs.get(output_var, {}))

            if 'shape' not in _expr_kwargs[output_var]:
                if common_shape is not _unspecified:
                    _expr_kwargs[output_var]['shape'] = (num_nodes,) + common_shape
                else:
                    _expr_kwargs[output_var]['shape'] = (num_nodes,)
            else:
                # Assume given shape is at a per-node basis
                _expr_kwargs[output_var]['shape'] = (num_nodes,) + _expr_kwargs[output_var]['shape']

            if 'units' not in _expr_kwargs[output_var]:
                if common_units is not _unspecified:
                    _expr_kwargs[output_var]['units'] = common_units

            scalar_src_idxs = []
            param_options = self._parameter_options or {}
            scalar_sources = list(param_options.keys()) + ['t_initial', 't_duration', 't_final']

            for rel_path, idx_str in re.findall(var_rgx, rhs):
                exec_var_name, src_idxs = parser.parse(rel_path, idx_str)
                expr = expr.replace(rel_path, exec_var_name)
                if '.' in rel_path:
                    self.connect(rel_path, exec_var_name, src_indices=src_idxs)

                # Only provide kwargs for things that we havent already done so.
                if exec_var_name not in seen_kwargs:
                    # Use deepcopy so we don't accidentally permanently set the shape here when we assign it.
                    _expr_kwargs[exec_var_name] = deepcopy(expr_kwargs.get(rel_path, {}))
                    if 'shape' not in _expr_kwargs[exec_var_name]:
                        if common_shape is not _unspecified:
                            _expr_kwargs[exec_var_name]['shape'] = (num_nodes,) + common_shape
                        else:
                            _expr_kwargs[exec_var_name]['shape'] = (num_nodes,)
                    else:
                        _expr_kwargs[exec_var_name]['shape'] = (num_nodes,) + _expr_kwargs[exec_var_name]['shape']

                    if exec_var_name in scalar_sources:
                        scalar_src_idxs.append(exec_var_name)

            seen_kwargs |= _expr_kwargs.keys()
            ec.add_expr(expr, **_expr_kwargs)
            if scalar_src_idxs:
                self.promotes('exec_comp', inputs=scalar_src_idxs,
                              src_indices=np.zeros(num_nodes, dtype=int))


def _make_ode_system(ode_class, num_nodes, ode_init_kwargs=None, calc_exprs=None, parameter_options=None):
    """
    Instantiate the ODE system, optionally including an ExecComp.

    Parameters
    ----------
    ode_class : System class
        The ODE class provided to the dymos phase.
    num_nodes : int
        The number of nodes at which the ODE is instantiated.
    ode_init_kwargs : dict or None
        Additional arguments we need to instantiate the ODE.
    calc_exprs : dict
        A dictionary keyed by exec comp expresions whose associated values
        are the metadata associated with the expression.
    parameter_options : dict
        The parameter options of the owning phase.

    Returns
    -------
    ode : System
        The instantiation of ode_class.  If `calc_exprs` were given, it
        returns a group that wraps the instantiated ode_class, as well
        as an exec comp used to evaluate the expressions.
    """
    _kwargs = ode_init_kwargs or {}
    if not calc_exprs:
        return ode_class(num_nodes=num_nodes, **_kwargs)
    else:
        ode_group = ODEGroup(ode_class,
                             num_nodes=num_nodes,
                             ode_init_kwargs=ode_init_kwargs,
                             calc_exprs=calc_exprs,
                             parameter_options=parameter_options)
        return ode_group



import openmdao.api as om
import numpy as np


class TimeAdder(om.ExplicitComponent):
    # This serves to keep track of the elapsed time. Since we performed a change of variables we
    # need to integrate time by adding dt/ds

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # states
        self.add_input('sdot', val=np.zeros(nn), desc='distance along track', units='m/s')

        # outputs
        self.add_output('dt_ds', val=np.zeros(nn), desc='distance perpendicular to centerline',
                        units='s/m')

        # Setup partials
        arange = np.arange(self.options['num_nodes'], dtype=int)

        # partials
        self.declare_partials(of='dt_ds', wrt='sdot', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        sdot = inputs['sdot']

        outputs['dt_ds'] = 1/sdot

    def compute_partials(self, inputs, jacobian):
        sdot = inputs['sdot']

        jacobian['dt_ds', 'sdot'] = -1/sdot**2



from __future__ import print_function, division, absolute_import

import numpy as np

from openmdao.api import ExplicitComponent


class ArcLengthComp(ExplicitComponent):

    def initialize(self):

        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('x', val=np.ones(nn), units='m', desc='x at points along the trajectory')
        self.add_input('theta', val=np.ones(nn), units='rad',
                       desc='wire angle with vertical along the trajectory')

        self.add_output('S', val=1.0, units='m', desc='arclength of wire')

        self.declare_partials(of='S', wrt='*', method='cs')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        x = inputs['x']
        theta = inputs['theta']

        dy_dx = -1.0 / np.tan(theta)
        dx = np.diff(x)
        f = np.sqrt(1 + dy_dx**2)

        # trapezoidal rule
        fxm1 = f[:-1]
        fx = f[1:]
        outputs['S'] = 0.5 * np.dot(fxm1 + fx, dx)



"""
Simple model for a set of motors in parallel where efficiency is a function of current.
"""
import numpy as np

import openmdao.api as om


class Motors(om.ExplicitComponent):
    """
    Model for motors in parallel.
    """

    def initialize(self):
        self.options.declare('num_nodes', default=1)
        self.options.declare('n_parallel', default=3, desc='number of motors in parallel')

    def setup(self):
        num_nodes = self.options['num_nodes']

        # Inputs
        self.add_input('power_out_gearbox', val=3.6*np.ones(num_nodes), units='W',
                       desc='Power at gearbox output')
        self.add_input('current_in_motor', val=np.ones(num_nodes), units='A',
                       desc='Total current demanded')

        # Outputs
        self.add_output('power_in_motor', val=np.ones(num_nodes), units='W',
                        desc='Power required at motor input')

        # Derivatives
        row_col = np.arange(num_nodes)

        self.declare_partials(of='power_in_motor', wrt=['*'], rows=row_col, cols=row_col)

    def compute(self, inputs, outputs):
        current = inputs['current_in_motor']
        power_out = inputs['power_out_gearbox']
        n_parallel = self.options['n_parallel']

        # Simple linear curve fit for efficiency.
        eff = 0.9 - 0.3 * current / n_parallel

        outputs['power_in_motor'] = power_out / eff

    def compute_partials(self, inputs, partials):
        current = inputs['current_in_motor']
        power_out = inputs['power_out_gearbox']
        n_parallel = self.options['n_parallel']

        eff = 0.9 - 0.3 * current / n_parallel

        partials['power_in_motor', 'power_out_gearbox'] = 1.0 / eff
        partials['power_in_motor', 'current_in_motor'] = 0.3 * power_out / (n_parallel * eff**2)


if __name__ == '__main__':

    import openmdao.api as om
    num_nodes = 1

    prob = om.Problem(model=Motors(num_nodes=num_nodes))
    model = prob.model

    prob.setup()

    prob.run_model()

    derivs = prob.check_partials(compact_print=True)

    print('done')



import numpy as np
import scipy.sparse as sp
from scipy.linalg import block_diag

import openmdao.api as om

from ..grid_data import GridData
from ...utils.misc import get_rate_units, CoerceDesvar
from ...utils.lgl import lgl
from ...utils.lagrange import lagrange_matrices
from ...utils.indexing import get_desvar_indices
from ...utils.constants import INF_BOUND
from ..._options import options as dymos_options


class ControlInterpComp(om.ExplicitComponent):
    """
    Class definition for the ControlInterpComp.

    Compute the approximated control values and rates given the values of a control at all nodes,
    given values at the control discretization nodes.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    Notes
    -----
    .. math::

        u = \\left[ L \\right] u_d

        \\dot{u} = \\frac{d\\tau_s}{dt} \\left[ D \\right] u_d

        \\ddot{u} = \\left( \\frac{d\\tau_s}{dt} \\right)^2 \\left[ D_2 \\right] u_d

    where
    :math:`u_d` are the values of the control at the control discretization nodes,
    :math:`u` are the values of the control at all nodes,
    :math:`\\dot{u}` are the time-derivatives of the control at all nodes,
    :math:`\\ddot{u}` are the second time-derivatives of the control at all nodes,
    :math:`L` is the Lagrange interpolation matrix,
    :math:`D` is the Lagrange differentiation matrix,
    and :math:`\\frac{d\\tau_s}{dt}` is the ratio of segment duration in segment tau space
    [-1 1] to segment duration in time.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'control_options', types=dict,
            desc='Dictionary of options for the dynamic controls')
        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info for the control inputs.')
        self.options.declare('output_grid_data', types=GridData, allow_none=True, default=None,
                             desc='GridData object for the output grid. If None, use the same grid_data as the inputs.')

        # Save the names of the dynamic controls/parameters
        self._input_names = {}
        self._output_val_names = {}
        self._output_rate_names = {}
        self._output_rate2_names = {}
        self._matrices = {}

    def setup(self):
        """
        Perform setup procedure for the Control interpolation component.
        """
        gd = self.options['grid_data']
        ogd = self.options['output_grid_data'] or gd

        if not gd.is_aligned_with(ogd):
            raise RuntimeError(f'{self.pathname}: The input grid and the output grid must have the same number of '
                               f'segments and segment spacing, but the input grid segment ends are '
                               f'\n{gd.segment_ends}\n and the output grid segment ends are \n'
                               f'{ogd.segment_ends}.')

    def _configure_controls(self):
        gd = self.options['grid_data']
        ogd = self.options['output_grid_data'] or gd
        eval_nodes = ogd.node_ptau
        control_options = self.options['control_options']
        num_output_nodes = ogd.num_nodes
        time_units = self.options['time_units']

        for name, options in control_options.items():
            if 'control_type' not in options:
                options['control_type'] = 'full'
            if options['control_type'] == 'polynomial':
                disc_nodes, _ = lgl(options['order'] + 1)
                num_control_input_nodes = len(disc_nodes)
                shape = options['shape']
                size = np.prod(shape)
                units = options['units']
                rate_units = get_rate_units(units, self.options['time_units'], deriv=1)
                rate2_units = get_rate_units(units, self.options['time_units'], deriv=2)

                input_shape = (num_control_input_nodes,) + shape
                output_shape = (num_output_nodes,) + shape

                L_de, D_de = lagrange_matrices(disc_nodes, eval_nodes)
                _, D_dd = lagrange_matrices(disc_nodes, disc_nodes)
                D2_de = np.dot(D_de, D_dd)

                self._matrices[name] = L_de, D_de, D2_de

                self._input_names[name] = f'controls:{name}'
                self._output_val_names[name] = f'control_values:{name}'
                self._output_rate_names[name] = f'control_rates:{name}_rate'
                self._output_rate2_names[name] = f'control_rates:{name}_rate2'

                self.add_input(self._input_names[name], val=np.ones(input_shape), units=units)
                self.add_output(self._output_val_names[name], shape=output_shape, units=units)
                self.add_output(self._output_rate_names[name], shape=output_shape, units=rate_units)
                self.add_output(self._output_rate2_names[name], shape=output_shape, units=rate2_units)

                self.val_jacs[name] = np.zeros((num_output_nodes, size, num_control_input_nodes, size))
                self.rate_jacs[name] = np.zeros((num_output_nodes, size, num_control_input_nodes, size))
                self.rate2_jacs[name] = np.zeros((num_output_nodes, size, num_control_input_nodes, size))

                for i in range(size):
                    self.val_jacs[name][:, i, :, i] = L_de
                    self.rate_jacs[name][:, i, :, i] = D_de
                    self.rate2_jacs[name][:, i, :, i] = D2_de

                self.val_jacs[name] = self.val_jacs[name].reshape((num_output_nodes * size,
                                                                  num_control_input_nodes * size),
                                                                  order='C')
                self.rate_jacs[name] = self.rate_jacs[name].reshape((num_output_nodes * size,
                                                                    num_control_input_nodes * size),
                                                                    order='C')
                self.rate2_jacs[name] = self.rate2_jacs[name].reshape((num_output_nodes * size,
                                                                      num_control_input_nodes * size),
                                                                      order='C')
                self.val_jac_rows[name], self.val_jac_cols[name] = \
                    np.where(self.val_jacs[name] != 0)
                self.rate_jac_rows[name], self.rate_jac_cols[name] = \
                    np.where(self.rate_jacs[name] != 0)
                self.rate2_jac_rows[name], self.rate2_jac_cols[name] = \
                    np.where(self.rate2_jacs[name] != 0)

                self.sizes[name] = size

                rs, cs = self.val_jac_rows[name], self.val_jac_cols[name]
                self.declare_partials(of=self._output_val_names[name],
                                      wrt=self._input_names[name],
                                      rows=rs, cols=cs, val=self.val_jacs[name][rs, cs])

                rs = np.concatenate([np.arange(0, num_output_nodes * size, size, dtype=int) + i
                                    for i in range(size)])

                self.declare_partials(of=self._output_rate_names[name],
                                      wrt='t_duration', rows=rs, cols=np.zeros_like(rs))

                self.declare_partials(of=self._output_rate_names[name],
                                      wrt=self._input_names[name],
                                      rows=self.rate_jac_rows[name], cols=self.rate_jac_cols[name])

                self.declare_partials(of=self._output_rate2_names[name],
                                      wrt='t_duration', rows=rs, cols=np.zeros_like(rs))

                self.declare_partials(of=self._output_rate2_names[name],
                                      wrt=self._input_names[name],
                                      rows=self.rate2_jac_rows[name], cols=self.rate2_jac_cols[name])

            else:
                self._input_names[name] = f'controls:{name}'
                self._output_val_names[name] = f'control_values:{name}'
                self._output_rate_names[name] = f'control_rates:{name}_rate'
                self._output_rate2_names[name] = f'control_rates:{name}_rate2'
                shape = options['shape']
                num_control_input_nodes = gd.subset_num_nodes['control_input']
                input_shape = (num_control_input_nodes,) + shape
                output_shape = (num_output_nodes,) + shape

                units = options['units']
                rate_units = get_rate_units(units, time_units)
                rate2_units = get_rate_units(units, time_units, deriv=2)

                self.add_input(self._input_names[name], val=np.ones(input_shape), units=units)

                self.add_output(self._output_val_names[name], shape=output_shape, units=units)

                self.add_output(self._output_rate_names[name], shape=output_shape, units=rate_units)

                self.add_output(self._output_rate2_names[name], shape=output_shape,
                                units=rate2_units)

                size = np.prod(shape)
                self.sizes[name] = size
                sp_eye = sp.eye(size, format='csr')

                # The partial of interpolated value wrt the control input values is linear
                # and can be computed as the kronecker product of the interpolation matrix (L)
                # and eye(size).
                J_val = sp.kron(self.L, sp_eye, format='csr')
                rs, cs, data = sp.find(J_val)
                self.declare_partials(of=self._output_val_names[name],
                                      wrt=self._input_names[name],
                                      rows=rs, cols=cs, val=data)

                # The partials of the output rate and second derivative wrt dt_dstau
                rs = np.arange(num_output_nodes * size, dtype=int)
                cs = np.repeat(np.arange(num_output_nodes, dtype=int), size)

                self.declare_partials(of=self._output_rate_names[name],
                                      wrt='dt_dstau',
                                      rows=rs, cols=cs)

                self.declare_partials(of=self._output_rate2_names[name],
                                      wrt='dt_dstau',
                                      rows=rs, cols=cs)

                # The partials of the rates and second derivatives are nonlinear but the sparsity
                # pattern is obtained from the kronecker product of the 1st and 2nd differentiation
                # matrices (D and D2) and eye(size).
                self.rate_jacs[name] = sp.kron(self.D, sp_eye, format='csr')
                rs, cs = self.rate_jacs[name].nonzero()

                self.declare_partials(of=self._output_rate_names[name],
                                      wrt=self._input_names[name],
                                      rows=rs, cols=cs)

                self.rate2_jacs[name] = sp.kron(self.D2, sp_eye, format='csr')
                rs, cs = self.rate2_jacs[name].nonzero()

                self.declare_partials(of=self._output_rate2_names[name],
                                      wrt=self._input_names[name],
                                      rows=rs, cols=cs)

    def _configure_desvars(self):
        control_options = self.options['control_options']
        gd = self.options['grid_data']

        for name, options in control_options.items():
            if options['control_type'] == 'polynomial':
                num_input_nodes = options['order'] + 1
                shape = options['shape']
                # default_val = reshape_val(options['val'], shape, num_input_nodes)
                if options['opt']:

                    desvar_indices = np.arange(num_input_nodes, dtype=int)
                    if options['fix_initial']:
                        desvar_indices = desvar_indices[1:]
                    if options['fix_final']:
                        desvar_indices = desvar_indices[:-1]

                    lb = -INF_BOUND if options['lower'] is None else options['lower']
                    ub = INF_BOUND if options['upper'] is None else options['upper']

                    self.add_design_var(f'controls:{name}',
                                        lower=lb,
                                        upper=ub,
                                        ref=options['ref'],
                                        ref0=options['ref0'],
                                        adder=options['adder'],
                                        scaler=options['scaler'],
                                        indices=desvar_indices,
                                        flat_indices=True)

                # self.set_input_defaults(name=f'controls:{name}', val=default_val, units=options['units'])
            else:
                num_input_nodes = gd.subset_num_nodes['control_input']

                dvname = f'controls:{name}'
                shape = options['shape']
                size = np.prod(shape)
                if options['opt']:
                    desvar_indices = get_desvar_indices(size, num_input_nodes,
                                                        options['fix_initial'], options['fix_final'])

                    if len(desvar_indices) > 0:
                        coerce_desvar_option = CoerceDesvar(num_input_nodes, desvar_indices,
                                                            options=options)

                        lb = np.zeros_like(desvar_indices, dtype=float)
                        lb[:] = -INF_BOUND if coerce_desvar_option('lower') is None else \
                            coerce_desvar_option('lower')

                        ub = np.zeros_like(desvar_indices, dtype=float)
                        ub[:] = INF_BOUND if coerce_desvar_option('upper') is None else \
                            coerce_desvar_option('upper')

                        self.add_design_var(name=dvname,
                                            lower=lb,
                                            upper=ub,
                                            scaler=coerce_desvar_option('scaler'),
                                            adder=coerce_desvar_option('adder'),
                                            ref0=coerce_desvar_option('ref0'),
                                            ref=coerce_desvar_option('ref'),
                                            indices=desvar_indices,
                                            flat_indices=True)

                # default_val = reshape_val(options['val'], shape, num_input_nodes)

                # self.set_input_defaults(name=dvname, val=default_val, units=options['units'])

    def _configure_desvars(self):
        control_options = self.options['control_options']
        gd = self.options['grid_data']

        for name, options in control_options.items():
            if options['control_type'] == 'polynomial':
                num_input_nodes = options['order'] + 1
                shape = options['shape']
                # default_val = reshape_val(options['val'], shape, num_input_nodes)
                if options['opt']:

                    desvar_indices = np.arange(num_input_nodes, dtype=int)
                    if options['fix_initial']:
                        desvar_indices = desvar_indices[1:]
                    if options['fix_final']:
                        desvar_indices = desvar_indices[:-1]

                    lb = -INF_BOUND if options['lower'] is None else options['lower']
                    ub = INF_BOUND if options['upper'] is None else options['upper']

                    self.add_design_var(f'controls:{name}',
                                        lower=lb,
                                        upper=ub,
                                        ref=options['ref'],
                                        ref0=options['ref0'],
                                        adder=options['adder'],
                                        scaler=options['scaler'],
                                        indices=desvar_indices,
                                        flat_indices=True)

                # self.set_input_defaults(name=f'controls:{name}', val=default_val, units=options['units'])
            else:
                num_input_nodes = gd.subset_num_nodes['control_input']

                dvname = f'controls:{name}'
                shape = options['shape']
                size = np.prod(shape)
                if options['opt']:
                    desvar_indices = get_desvar_indices(size, num_input_nodes,
                                                        options['fix_initial'], options['fix_final'])

                    if len(desvar_indices) > 0:
                        coerce_desvar_option = CoerceDesvar(num_input_nodes, desvar_indices,
                                                            options=options)

                        lb = np.zeros_like(desvar_indices, dtype=float)
                        lb[:] = -INF_BOUND if coerce_desvar_option('lower') is None else \
                            coerce_desvar_option('lower')

                        ub = np.zeros_like(desvar_indices, dtype=float)
                        ub[:] = INF_BOUND if coerce_desvar_option('upper') is None else \
                            coerce_desvar_option('upper')

                        self.add_design_var(name=dvname,
                                            lower=lb,
                                            upper=ub,
                                            scaler=coerce_desvar_option('scaler'),
                                            adder=coerce_desvar_option('adder'),
                                            ref0=coerce_desvar_option('ref0'),
                                            ref=coerce_desvar_option('ref'),
                                            indices=desvar_indices,
                                            flat_indices=True)

                # default_val = reshape_val(options['val'], shape, num_input_nodes)

                # self.set_input_defaults(name=dvname, val=default_val, units=options['units'])

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the states.
        """
        time_units = self.options['time_units']
        gd = self.options['grid_data']
        ogd = self.options['output_grid_data'] or gd
        output_num_nodes = ogd.subset_num_nodes['all']
        num_seg = gd.num_segments

        self.add_input('dt_dstau', shape=output_num_nodes, units=time_units)

        self.val_jacs = {}
        self.rate_jacs = {}
        self.rate2_jacs = {}
        self.val_jac_rows = {}
        self.val_jac_cols = {}
        self.rate_jac_rows = {}
        self.rate_jac_cols = {}
        self.rate2_jac_rows = {}
        self.rate2_jac_cols = {}
        self.sizes = {}

        self.add_input('t_duration', val=1.0, units=self.options['time_units'],
                       desc='duration of the phase to which this interpolated control group '
                            'belongs')

        num_disc_nodes = gd.subset_num_nodes['control_disc']
        num_input_nodes = gd.subset_num_nodes['control_input']

        # Find the indexing matrix that, multiplied by the values at the input nodes,
        # gives the values at the discretization nodes
        L_id = np.zeros((num_disc_nodes, num_input_nodes), dtype=float)
        L_id[np.arange(num_disc_nodes, dtype=int),
             gd.input_maps['dynamic_control_input_to_disc']] = 1.0
        L_id = sp.csr_matrix(L_id)

        # Matrices L_da and D_da interpolate values and rates (respectively) at all nodes from
        # values specified at control discretization nodes.
        # If the output grid is different than the input grid, then we have to build these these matrices ourselves.
        if ogd is gd:
            L_da, D_da = gd.phase_lagrange_matrices('control_disc', 'all', sparse=True)
        else:
            L_blocks = []
            D_blocks = []

            for iseg in range(num_seg):
                i1, i2 = gd.subset_segment_indices['control_disc'][iseg, :]
                indices = gd.subset_node_indices['control_disc'][i1:i2]
                nodes_given = gd.node_stau[indices]

                i1, i2 = ogd.subset_segment_indices['all'][iseg, :]
                indices = ogd.subset_node_indices['all'][i1:i2]
                nodes_eval = ogd.node_stau[indices]

                L_block, D_block = lagrange_matrices(nodes_given, nodes_eval)

                L_blocks.append(L_block)
                D_blocks.append(D_block)

            L_da = sp.csr_matrix(block_diag(*L_blocks))
            D_da = sp.csr_matrix(block_diag(*D_blocks))

        self.L = L_da.dot(L_id)
        self.D = D_da.dot(L_id)

        # Matrix D_dd interpolates rates at discretization nodes from values given at control
        # discretization nodes.
        _, D_dd = gd.phase_lagrange_matrices('control_disc', 'control_disc', sparse=True)

        # Matrix D2 provides second derivatives at all nodes given values at input nodes.
        self.D2 = D_da.dot(D_dd.dot(L_id))

        self._configure_controls()
        self._configure_desvars()

    def compute(self, inputs, outputs):
        """
        Compute interpolated control values and rates.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        gd = self.options['grid_data']
        ogd = self.options['output_grid_data'] or gd
        control_options = self.options['control_options']
        num_control_input_nodes = self.options['grid_data'].subset_num_nodes['control_input']
        num_output_nodes = ogd.num_nodes
        dt_dptau = 0.5 * inputs['t_duration']

        for name, options in control_options.items():
            if control_options[name]['control_type'] == 'polynomial':
                L_de, D_de, D2_de = self._matrices[name]

                u = inputs[self._input_names[name]]

                a = np.tensordot(D_de, u, axes=(1, 0)).T
                b = np.tensordot(D2_de, u, axes=(1, 0)).T

                # divide each "row" by dt_dptau or dt_dptau**2
                outputs[self._output_val_names[name]] = np.tensordot(L_de, u, axes=(1, 0))
                outputs[self._output_rate_names[name]] = (a / dt_dptau).T
                outputs[self._output_rate2_names[name]] = (b / dt_dptau ** 2).T
            else:
                size = np.prod(options['shape'])

                u_flat = np.reshape(inputs[self._input_names[name]],
                                    (num_control_input_nodes, size))

                a = self.D.dot(u_flat)
                b = self.D2.dot(u_flat)

                val = np.reshape(self.L.dot(u_flat), (num_output_nodes,) + options['shape'])

                rate = a / inputs['dt_dstau'][:, np.newaxis]
                rate = np.reshape(rate, (num_output_nodes,) + options['shape'])

                rate2 = b / inputs['dt_dstau'][:, np.newaxis] ** 2
                rate2 = np.reshape(rate2, (num_output_nodes,) + options['shape'])

                outputs[self._output_val_names[name]] = val
                outputs[self._output_rate_names[name]] = rate
                outputs[self._output_rate2_names[name]] = rate2

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        ogd = self.options['output_grid_data'] or self.options['grid_data']
        control_options = self.options['control_options']

        dstau_dt = np.reciprocal(inputs['dt_dstau'])
        dstau_dt2 = (dstau_dt ** 2)[:, np.newaxis]
        dstau_dt3 = (dstau_dt ** 3)[:, np.newaxis]

        t_duration = inputs['t_duration']

        for name in control_options:
            if control_options[name]['control_type'] == 'polynomial':
                options = control_options[name]
                control_name = self._input_names[name]
                num_input_nodes = options['order'] + 1
                L_de, D_de, D2_de = self._matrices[name]
                nn = ogd.num_nodes

                size = self.sizes[name]
                rate_name = self._output_rate_names[name]
                rate2_name = self._output_rate2_names[name]

                # Unroll matrix-shaped controls into an array at each node
                u_d = np.reshape(inputs[control_name], (num_input_nodes, size))

                t_duration_tile = np.tile(t_duration, size * nn)

                partials[rate_name, 't_duration'] = \
                    0.5 * (-np.dot(D_de, u_d).ravel(order='F') / (0.5 * t_duration_tile) ** 2)

                partials[rate2_name, 't_duration'] = \
                    -1.0 * (np.dot(D2_de, u_d).ravel(order='F') / (0.5 * t_duration_tile) ** 3)

                t_duration_x_size = np.repeat(t_duration, size * nn)[:, np.newaxis]

                r_nz, c_nz = self.rate_jac_rows[name], self.rate_jac_cols[name]
                partials[rate_name, control_name] = \
                    (self.rate_jacs[name] / (0.5 * t_duration_x_size))[r_nz, c_nz]

                r_nz, c_nz = self.rate2_jac_rows[name], self.rate2_jac_cols[name]
                partials[rate2_name, control_name] = \
                    (self.rate2_jacs[name] / (0.5 * t_duration_x_size) ** 2)[r_nz, c_nz]
            else:
                control_name = self._input_names[name]

                size = self.sizes[name]
                rate_name = self._output_rate_names[name]
                rate2_name = self._output_rate2_names[name]
                num_input_nodes = self.options['grid_data'].subset_num_nodes['control_input']

                # Unroll shaped controls into an array at each node
                u_flat = np.reshape(inputs[control_name], (num_input_nodes, size))

                partials[rate_name, 'dt_dstau'] = (-self.D.dot(u_flat) * dstau_dt2).ravel()
                partials[rate2_name, 'dt_dstau'] = (-2.0 * self.D2.dot(u_flat) * dstau_dt3).ravel()

                dstau_dt_x_size = np.repeat(dstau_dt, size)[:, np.newaxis]
                dstau_dt2_x_size = np.repeat(dstau_dt2, size)[:, np.newaxis]

                partials[rate_name, control_name] = self.rate_jacs[name].multiply(dstau_dt_x_size).data

                partials[rate2_name, control_name] = self.rate2_jacs[name].multiply(dstau_dt2_x_size).data



import functools
import numpy as np

from scipy.linalg import block_diag
import scipy.sparse as sp

from dymos.utils.lgl import lgl
from dymos.utils.lgr import lgr
from dymos.utils.cgl import cgl
from dymos.utils.hermite import hermite_matrices
from dymos.utils.lagrange import lagrange_matrices


def gauss_lobatto_subsets_and_nodes(n, seg_idx, compressed=False):
    """
    Provides node information and the location of the nodes for n Legendre-Gauss-Lobatto nodes on the range [-1, 1].

    Parameters
    ----------
    n : int
        The total number of nodes in the Gauss-Lobatto segment.  Must be
        an odd number.
    seg_idx : int
        The index of this segment within its phase.
    compressed : bool
        True if the subset requested is for a phase with compressed transcription.

    Returns
    -------
    dict
        A dictionary with the following keys:
        'state_disc' Gives the indices of the state discretization nodes
        'state_input' Gives the indices of the state input nodes
        'control_disc' Gives the indices of the control discretization nodes
        'control_input' Gives the indices of the control input nodes
        'segment_ends' Gives the indices of the nodes at the start (even) and end (odd) of a segment
        'col' Gives the indices of the collocation nodes.
        'all' Gives all node indices.
    np.array
        The location of all nodes on [-1, 1].
    np.array
        The weights of all nodes on [-1, 1].

    Notes
    -----
    Subset 'state_input' is the same as subset 'state_disc' if `compressed == False` or
    `first_seg == True`.  The same is true of subsets 'control_input' and 'control_disc'.
    """
    if n < 2:
        raise ValueError('The number of nodes must be larger than 1.')

    subsets = {
        'state_disc': np.arange(0, n, 2, dtype=int),
        'state_input': np.arange(0, n, 2, dtype=int) if not compressed or seg_idx == 0
        else np.arange(2, n, 2, dtype=int),
        'control_disc': np.arange(n, dtype=int),
        'control_input': np.arange(n, dtype=int) if not compressed or seg_idx == 0
        else np.arange(1, n, dtype=int),
        'segment_ends': np.array([0, n-1], dtype=int),
        'col': np.arange(1, n, 2, dtype=int),
        'all': np.arange(n, dtype=int),
        'solution': np.arange(n, dtype=int),
    }

    return subsets, *lgl(n)


def chebyshev_gauss_lobatto_subsets_and_nodes(n, seg_idx, compressed=False):
    """
    Provides node information and the location of the nodes for n Legendre-Gauss-Lobatto nodes on the range [-1, 1].

    Parameters
    ----------
    n : int
        The total number of nodes in the Gauss-Lobatto segment.  Must be
        an odd number.
    seg_idx : int
        The index of this segment within its phase.
    compressed : bool
        True if the subset requested is for a phase with compressed transcription.

    Returns
    -------
    dict
        A dictionary with the following keys:
        'state_disc' Gives the indices of the state discretization nodes
        'state_input' Gives the indices of the state input nodes
        'control_disc' Gives the indices of the control discretization nodes
        'control_input' Gives the indices of the control input nodes
        'segment_ends' Gives the indices of the nodes at the start (even) and end (odd) of a segment
        'col' Gives the indices of the collocation nodes.
        'all' Gives all node indices.
    np.array
        The location of all nodes on [-1, 1].
    np.array
        The weights of all nodes on [-1, 1].

    Notes
    -----
    Subset 'state_input' is the same as subset 'state_disc' if `compressed == False` or
    `first_seg == True`.  The same is true of subsets 'control_input' and 'control_disc'.
    """
    if n < 2:
        raise ValueError('The number of nodes must be larger than 1.')

    subsets = {
        'state_disc': np.arange(n, dtype=int),
        'state_input': np.arange(n, dtype=int) if not compressed or seg_idx == 0
        else np.arange(1, n, 2, dtype=int),
        'control_disc': np.arange(n, dtype=int),
        'control_input': np.arange(n, dtype=int) if not compressed or seg_idx == 0
        else np.arange(1, n, dtype=int),
        'segment_ends': np.array([0, n-1], dtype=int),
        'col': np.arange(n, dtype=int),
        'all': np.arange(n, dtype=int),
        'solution': np.arange(n, dtype=int),
    }

    return subsets, *cgl(n)


def radau_pseudospectral_subsets_and_nodes(n, seg_idx, compressed=False):
    """
    Provides node information and the location of the nodes for n Radau nodes on the range [-1, 1].

    Parameters
    ----------
    n : int
        The total number of nodes in the Radau Pseudospectral segment (including right endpoint).
    seg_idx : int
        The index of this segment within its phase.
    compressed : bool
        True if the subset requested is for a phase with compressed transcription.

    Returns
    -------
    dict
        A dictionary with the following keys:
        'state_disc' gives the indices of the state discretization nodes.
        'state_input' gives the indices of the state input nodes.
        'control_disc' gives the indices of the control discretization nodes.
        'control_input' gives the indices of the control input nodes.
        'segment_ends' gives the indices of the nodes at the start (even) and end (odd) of a segment.
        'col' gives the indices of the collocation nodes.
        'all' gives all node indices.
    np.array
        The location of all nodes on [-1, 1].
    np.array
        The weights of all nodes on [-1, 1].

    Notes
    -----
    Subset 'state_input' is the same as subset 'state_disc' if `compressed == False` or
    `first_seg == True`.  For Radau-Pseudospectral transcription, subset 'control_input' is always
    the same as subset 'control_disc'.
    """
    subsets = {
        'state_disc': np.arange(n + 1, dtype=int),
        'state_input': np.arange(n + 1, dtype=int) if not compressed or seg_idx == 0
        else np.arange(1, n + 1, dtype=int),
        'control_disc': np.arange(n, dtype=int),
        'control_input': np.arange(n, dtype=int),
        'segment_ends': np.array([0, n], dtype=int),
        'col': np.arange(n, dtype=int),
        'all': np.arange(n + 1, dtype=int),
        'solution': np.arange(n + 1, dtype=int),
    }

    return subsets, *lgr(n, include_endpoint=True)


def birkhoff_subsets_and_nodes(n, grid, seg_idx, compressed=False):
    """
    Provides node information and the location of the nodes for n Radau nodes on the range [-1, 1].

    Parameters
    ----------
    n : int
        The total number of nodes in the segment.
    grid : str
        The type of Gaussian grid used in the transcription.
    seg_idx : int
        The index of this segment within its phase.
    compressed : bool
        True if the subset requested is for a phase with compressed transcription.

    Returns
    -------
    dict
        A dictionary with the following keys:
        'state_disc' gives the indices of the state discretization nodes.
        'state_input' gives the indices of the state input nodes.
        'control_disc' gives the indices of the control discretization nodes.
        'control_input' gives the indices of the control input nodes.
        'segment_ends' gives the indices of the nodes at the start (even) and end (odd) of a segment.
        'col' gives the indices of the collocation nodes.
        'all' gives all node indices.
    np.array
        The location of all nodes on [-1, 1].
    np.array
        The weights of all nodes on [-1, 1].

    Notes
    -----
    Subset 'state_input' is the same as subset 'state_disc' if `compressed == False` or
    `first_seg == True`.  For Radau-Pseudospectral transcription, subset 'control_input' is always
    the same as subset 'control_disc'.
    """
    acceptable_grids = {'lgl', 'lgr', 'cgl'}

    subsets = {
        'state_disc': np.arange(n, dtype=int),
        'state_input': np.arange(n, dtype=int),
        'control_disc': np.arange(n, dtype=int),
        'control_input': np.arange(n, dtype=int),
        'segment_ends': np.array([0, n], dtype=int),
        'col': np.arange(n, dtype=int),
        'solution': np.arange(n, dtype=int),
    }

    if grid == 'lgl':
        nodes, weights = lgl(n)
        subsets['all'] = np.arange(n, dtype=int)
    elif grid == 'lgr':
        nodes, weights = lgr(n, include_endpoint=False)
        subsets['all'] = np.arange(n, dtype=int)
    elif grid == 'cgl':
        nodes, weights = cgl(n)
        subsets['all'] = np.arange(n, dtype=int)
    else:
        raise ValueError(f'Unrecognized grid. Acceptable values are one of {acceptable_grids}')

    return subsets, nodes, weights


def uniform_subsets_and_nodes(n, *args, **kwargs):
    """
    Provides a dict of node info and locations for a uniformly distributed set of n nodes on the range [-1, 1].

    This distribution is not to be used to define polynomials, since equally-spaced nodes
    result in poor polynomial fitting. Most subsets here aside from `all`, `segment_ends`, and
    `solution` are not defined for the uniform distribution.

    Parameters
    ----------
    n : int
        The total number of nodes in the Radau Pseudospectral segment (including right endpoint).
    *args : Iterable
        Non-keyword arguments that make this function consistent with radau_subsets_and_nodes and
        gauss_lobatto_subsets_and_nodes, but whose additional arguments have no bearing on the uniform grid.
    **kwargs : dict
        Keyword arguments that make this function consistent with radau_subsets_and_nodes and
        gauss_lobatto_subsets_and_nodes, but whose keyword arguments have no bearing on the uniform grid.

    Returns
    -------
    dict
        A dictionary with the following keys:
        'state_disc' gives the indices of the state discretization nodes.
        'state_input' gives the indices of the state input nodes.
        'control_disc' gives the indices of the control discretization nodes.
        'control_input' gives the indices of the control input nodes.
        'segment_ends' gives the indices of the nodes at the start (even) and end (odd) of a segment.
        'col' gives the indices of the collocation nodes.
        'all' gives all node indices.
    np.array
        The location of all nodes on [-1, 1].
    """
    subsets = {
        'state_disc': np.empty(0, dtype=int),
        'state_input': np.empty(0, dtype=int),
        'control_disc': np.empty(0, dtype=int),
        'control_input': np.empty(0, dtype=int),
        'segment_ends': np.array([0, n], dtype=int),
        'col': np.empty(0, dtype=int),
        'all': np.arange(n + 1, dtype=int),
        'solution': np.arange(n + 1, dtype=int),
    }
    weights = np.ones(n + 1)
    weights[1:-1] = 2.0
    return subsets, np.linspace(-1, 1, n + 1), weights


def make_subset_map(from_subset_idxs, to_subset_idxs):
    """
    Creates a map from one subset to another using the indices of each subset within all nodes.

    Parameters
    ----------
    from_subset_idxs : iterable of int
        Subset indices for the subset from which we are mapping values.
    to_subset_idxs : iterable of int
        Subset indices for the subset to which we are mapping values.

    Returns
    -------
    numpy.array of int
        An index map which, when applied to values in the from_subset, will provide values
        in the to_subset.
    """
    offset = 0
    subset_map = []
    for i in range(len(to_subset_idxs)):
        if to_subset_idxs[i] not in from_subset_idxs:
            offset += 1
        subset_map.append(i - offset)
    return np.array(subset_map)


class GridData(object):
    """
    Properties associated with the GridData of a phase.

    GridData contains properties associated
    with the "grid" or "mesh" of a phase - the number of segments, the
    polynomial order of each segment, and the relative lengths of the segments.
    In turn, these three defining properties determine various other properties,
    such as indexing arrays used to extract the discretization or collocation
    nodes from a list of all nodes within the phase.

    Parameters
    ----------
    num_segments : int
        The number of segments in the phase.
    transcription : str
        Case-insensitive distribution (e.g., ('gauss-lobatto', 'radau-ps', 'explicit', 'uniform')).
    transcription_order : int or int ndarray[:] or str
        The order of the state transcription in each segment, as a scalar or a vector.
    segment_ends : Iterable[num_segments + 1] or None
        The segments nodes on some arbitrary interval.
        This will be normalized to the interval [-1, 1].
    compressed : bool
        If the transcription is compressed, then states and controls at shared
        nodes of adjacent segments are only specified once, and then broadcast
        to the appropriate indices.
    num_steps_per_segment : int or None
        The number of steps to take in each segment, for explicitly integrated phases.

    Attributes
    ----------
    transcription : str
        The transcription to which this GridData instance applies.  One of
        'gauss-lobatto' or 'radau-ps'.
    num_segments : int
        The number of segments in the phase
    segment_ends : ndarray or None
        The segment boundaries in non-dimensional phase time (phase tau space).
        If given as a Iterable, it must be of length (num_segments+1) and it
        must be monotonically increasing.
        If None, then the segments are equally spaced.
    num_steps_per_segment : int
        The number of steps to take in each segment of the phase, for explicit phases.
    num_nodes : int
        The total number of nodes in the phase
    node_weight : ndarray
        The quadrature weight for each node on the interval [-1, 1].
    node_stau : ndarray
        The locations of each node in non-dimensional segment time (segment tau space).
    node_ptau : ndarray
        The locations of each node in non-dimensional phase time (phase tau space).
    node_dptau_dstau : ndarray
        The ratio of phase tau to segment tau at each node.
    segment_indices : int ndarray[:,2]
        Array where each row contains the start and end indices into the nodes.
    subset_node_indices : dict of int ndarray[:]
        Dict keyed by subset name where each entry are the indices of the nodes
        belonging to that given subset.
    subset_segment_indices: dict of int ndarray[num_seg,:]
        Dict keyed by subset name where each entry are the indices of the nodes
        belonging to the given subset, indexed into subset_node_indices!
    subset_num_nodes: dict of int
        A dict keyed by subset name that provides the total number of
        nodes in the phase which belong to the given subset.
    subset_num_nodes_per_segment: dict of list
        A dict keyed by subset name that provides a list of ints giving the number of
        nodes which belong to the given subset in each segment.
    compressed: bool
        True if the transcription is compressed (connecting nodes of adjacent segments
        are not duplicated in the inputs).
    input_maps: dict of int ndarray[:]
        Dict keyed by the map name that provides a mapping for src_indices to
        and from "compressed" form.
    """
    def __init__(self, num_segments, transcription, transcription_order=None,
                 segment_ends=None, compressed=False, num_steps_per_segment=1):
        if segment_ends is None:
            segment_ends = np.linspace(-1, 1, num_segments + 1)
        else:
            if len(segment_ends) != num_segments + 1:
                raise ValueError('segment_ends must be of length (num_segments + 1)')
            # Assert monotonic increasing
            if not np.all(np.diff(segment_ends) > 0):
                raise ValueError('segment_ends must be monotonically increasing')
            segment_ends = np.atleast_1d(segment_ends)

        v0 = segment_ends[0]
        v1 = segment_ends[-1]
        segment_ends = -1. + 2 * (segment_ends - v0) / (v1 - v0)

        # List of all GridData attributes

        self.num_segments = num_segments

        self.segment_ends = segment_ends

        self.num_nodes = 0

        self.num_steps_per_segment = 0

        self.node_weight = np.empty(0,)

        self.node_stau = np.empty(0,)

        self.node_ptau = np.empty(0,)

        self.node_dptau_dstau = np.empty(0,)

        self.segment_indices = np.empty((num_segments, 2), dtype=int)

        self.subset_node_indices = {}

        self.subset_segment_indices = {}

        self.subset_num_nodes = {}

        self.subset_num_nodes_per_segment = {}

        self.compressed = compressed

        self.input_maps = {'state_input_to_disc': np.empty(0, dtype=int),
                           'dynamic_control_input_to_disc': np.empty(0, dtype=int)}

        if transcription.lower() in ['radau', 'radau-ps', 'lgr']:
            self.transcription = 'radau-ps'
        elif transcription.lower() in ['gausslobatto', 'gauss-lobatto', 'lgl']:
            self.transcription = 'gauss-lobatto'
        elif transcription.lower() in ['chebyshev-gauss-lobatto', 'cgl']:
            self.transcription = 'chebyshev-gauss-lobatto'
        elif transcription.lower() in ['birkhoff']:
            self.transcription = 'birkhoff'
        elif transcription.lower() in ['uniform']:
            self.transcription = 'uniform'
        else:
            raise ValueError(f'Unknown transcription: {transcription}')

        # Define get_subsets and node points based on the transcription scheme
        if self.transcription == 'gauss-lobatto':
            get_subsets_and_nodes = gauss_lobatto_subsets_and_nodes
        if self.transcription == 'chebyshev-gauss-lobatto':
            get_subsets_and_nodes = chebyshev_gauss_lobatto_subsets_and_nodes
        elif self.transcription == 'radau-ps':
            get_subsets_and_nodes = radau_pseudospectral_subsets_and_nodes
        elif self.transcription == 'uniform':
            get_subsets_and_nodes = uniform_subsets_and_nodes
        elif self.transcription == 'birkhoff':
            get_subsets_and_nodes = functools.partial(birkhoff_subsets_and_nodes, grid=self.grid_type)

        # Make sure transcription_order is a vector
        if isinstance(transcription_order, str):
            self.transcription_order = num_segments * [transcription_order]
        elif np.ndim(transcription_order) == 0:  # scalar
            self.transcription_order = np.ones(num_segments, int) * transcription_order
        elif np.size(transcription_order) == 1:  # length-1 array
            self.transcription_order = np.ones(num_segments, int) * np.asarray(transcription_order)[0]
        else:
            self.transcription_order = np.asarray(transcription_order, dtype=int)

        # Make sure num_steps_per_segment is a vector
        num_steps_per_segment = np.ones(num_segments, int) * num_steps_per_segment
        self.num_steps_per_segment = num_steps_per_segment

        # Build up the arrays segment by segment
        self.segment_indices[0, 0] = 0
        ind0 = 0  # index of the first node in the segment
        for iseg in range(num_segments):
            subsets_i, nodes_i, weights_i = get_subsets_and_nodes(self.transcription_order[iseg],
                                                                  seg_idx=iseg,
                                                                  compressed=compressed)

            if iseg == 0:
                subset_ind0 = {name: 0 for name in subsets_i}
                # index of the first node in the segment for each subset
                subset_ind1 = subset_ind0.copy()
                # index of the last node in the segment for each subset

            num_nodes_i = len(nodes_i)
            self.num_nodes += num_nodes_i

            # Append our nodes in segment tau space
            self.node_stau = np.concatenate((self.node_stau, nodes_i))

            # Append our node weights
            self.node_weight = np.concatenate((self.node_weight, weights_i))

            # Append our nodes in phase tau space
            v0 = segment_ends[iseg]
            v1 = segment_ends[iseg + 1]
            self.node_ptau = np.concatenate((self.node_ptau, v0 + 0.5 * (nodes_i + 1) * (v1 - v0)))
            self.node_dptau_dstau = np.concatenate((self.node_dptau_dstau,
                                                    0.5 * (v1 - v0) * np.ones_like(nodes_i)))

            self.segment_indices[iseg, 1] = self.segment_indices[iseg, 0] + num_nodes_i
            if iseg < num_segments - 1:
                self.segment_indices[iseg + 1, 0] = self.segment_indices[iseg, 1]

            for subset_name, subset_idxs_i in subsets_i.items():
                if iseg == 0:
                    self.subset_num_nodes[subset_name] = 0
                    self.subset_num_nodes_per_segment[subset_name] = []
                    self.subset_segment_indices[subset_name] = np.zeros((num_segments, 2),
                                                                        dtype=int)
                    self.subset_node_indices[subset_name] = np.empty(0, dtype=int)

                if subset_idxs_i is None:
                    num_subset_nodes_i = 0
                else:
                    num_subset_nodes_i = len(subset_idxs_i)

                self.subset_num_nodes[subset_name] += num_subset_nodes_i
                self.subset_num_nodes_per_segment[subset_name].append(num_subset_nodes_i)
                subset_ind1[subset_name] += num_subset_nodes_i

                self.subset_segment_indices[subset_name][iseg, 0] = subset_ind0[subset_name]
                self.subset_segment_indices[subset_name][iseg, 1] = subset_ind1[subset_name]

                self.subset_node_indices[subset_name] = \
                    np.concatenate((self.subset_node_indices[subset_name], subset_idxs_i + ind0))

                subset_ind0[subset_name] += num_subset_nodes_i

            ind0 += num_nodes_i  # The first node in the next segment

        state_input_idxs = self.subset_node_indices['state_input']
        state_disc_idxs = self.subset_node_indices['state_disc']

        control_input_idxs = self.subset_node_indices['control_input']
        control_disc_idxs = self.subset_node_indices['control_disc']

        self.input_maps['state_input_to_disc'] = make_subset_map(state_input_idxs, state_disc_idxs)

        self.input_maps['dynamic_control_input_to_disc'] = make_subset_map(control_input_idxs,
                                                                           control_disc_idxs)

    def __eq__(self, other):
        """
        Compare this GridData with an object with other and return True if they are equivalent.

        Parameters
        ----------
        other : GridData

        Returns
        -------
        bool
            True if other is equivalent to self, otherwise False.
        """
        if isinstance(other, GridData):
            return self.transcription == other.transcription and \
                self.num_segments == other.num_segments and \
                np.all(self.segment_ends == other.segment_ends) and \
                self.compressed == other.compressed and \
                np.all(self.transcription_order == other.transcription_order) and \
                np.all(self.num_steps_per_segment == other.num_steps_per_segment)
        else:
            return False

    def __repr__(self):
        return f'{self.__class__.__name__}(num_seg={self.num_segments}, order={self.transcription_order}) at <{id(self)}>'

    def is_aligned_with(self, other, tol=1.0E-12):
        """
        Check that the segment distribution in GridData object `other` matches that of this GridData object.

        Parameters
        ----------
        other : GridData
            GridData object against which this one is being compared.
        tol : float
            The absolute tolerance in difference between segment end locations in phase tau space.

        Returns
        -------
        bool
            True if the two GridData objects have the same number of segments and their segments match to within
            the specified tolerance, otherwise False.
        """
        # The segment distribution needs to be the same in from the input grid to the output grid.
        return self.num_segments == other.num_segments and \
            (np.abs(self.segment_ends - other.segment_ends) <= tol).all()

    def phase_lagrange_matrices(self, given_set_name, eval_set_name, sparse=False):
        """
        Compute the matrices mapping values at some nodes to values and derivatives at new nodes.

        Parameters
        ----------
        given_set_name : str
            Name of the set of nodes with which to perform the interpolation.
        eval_set_name : str
            Name of the set of nodes at which to evaluate the values and derivatives.
        sparse : bool
            If True, the returned matrix will be in scipy CSR sparse format.  Otherwise, it is
            returned as a dense numpy.array.

        Returns
        -------
        ndarray[num_eval_set, num_given_set]
            Matrix containing the values at the new nodes.
        ndarray[num_eval_set, num_given_set]
            Matrix containing the time derivatives at the new nodes.

        Notes
        -----
        The values are mapped using the equation:

        .. math::

            x_{eval} = \\left[ L \\right] x_{given}

        And the derivatives are mapped with the equation:

        .. math::

            \\dot{x}_{eval} = \\left[ D \\right] x_{given} \\frac{d \\tau}{dt}
        """
        L_blocks = []
        D_blocks = []

        for iseg in range(self.num_segments):
            i1, i2 = self.subset_segment_indices[given_set_name][iseg, :]
            indices = self.subset_node_indices[given_set_name][i1:i2]
            nodes_given = self.node_stau[indices]

            i1, i2 = self.subset_segment_indices[eval_set_name][iseg, :]
            indices = self.subset_node_indices[eval_set_name][i1:i2]
            nodes_eval = self.node_stau[indices]

            L_block, D_block = lagrange_matrices(nodes_given, nodes_eval)

            L_blocks.append(L_block)
            D_blocks.append(D_block)

        L = block_diag(*L_blocks)
        D = block_diag(*D_blocks)

        if sparse:
            L = sp.csr_matrix(L)
            D = sp.csr_matrix(D)

        return L, D

    def phase_hermite_matrices(self, given_set_name, eval_set_name, sparse=False):
        """
        Compute the matrices mapping values at some nodes to values and derivatives at new nodes.

        Parameters
        ----------
        given_set_name : str
            Name of the set of nodes with which to perform the interpolation.
        eval_set_name : str
            Name of the set of nodes at which to evaluate the values and derivatives.
        sparse : bool
            If True, the returned matrix will be in scipy CSR sparse format.  Otherwise, it is
            returned as a dense numpy.array.

        Returns
        -------
        ndarray[num_eval_set, num_given_set]
            Matrix that maps values at given nodes to values at eval nodes.
            This is A_i in the equations above.
        ndarray[num_eval_set, num_given_set]
            Matrix that maps derivatives at given nodes to values at eval nodes.
            This is B_i in the equations above.
        ndarray[num_eval_set, num_given_set]
            Matrix that maps values at given nodes to derivatives at eval nodes.
            This is A_d in the equations above.
        ndarray[num_eval_set, num_given_set]
            Matrix that maps derivatives at given nodes to derivatives at eval nodes.
            This is A_d in the equations above.

        Notes
        -----
        The equation for Hermite interpolation of the values is:

        .. math::

            x_{eval} = \\left[ A_i \\right] x_{given}
                             + \\frac{dt}{d\\tau} \\left[ B_i \\right] \\dot{x}_{given}

        Hermite interpolation of the derivatives is performed as:

        .. math::

            \\dot{x}_{eval} = \\frac{d\\tau}{dt} \\left[ A_d \\right] x_{given}
                                   + \\left[ B_d \\right] \\dot{x}_{given}
        """
        Ai_list = []
        Bi_list = []
        Ad_list = []
        Bd_list = []

        for iseg in range(self.num_segments):
            i1, i2 = self.subset_segment_indices[given_set_name][iseg, :]
            indices = self.subset_node_indices[given_set_name][i1:i2]
            nodes_given = self.node_stau[indices]

            i1, i2 = self.subset_segment_indices[eval_set_name][iseg, :]
            indices = self.subset_node_indices[eval_set_name][i1:i2]
            nodes_eval = self.node_stau[indices]

            Ai_seg, Bi_seg, Ad_seg, Bd_seg = hermite_matrices(nodes_given, nodes_eval)

            Ai_list.append(Ai_seg)
            Bi_list.append(Bi_seg)
            Ad_list.append(Ad_seg)
            Bd_list.append(Bd_seg)

        Ai = block_diag(*Ai_list)
        Bi = block_diag(*Bi_list)
        Ad = block_diag(*Ad_list)
        Bd = block_diag(*Bd_list)

        if sparse:
            Ai = sp.csr_matrix(Ai)
            Bi = sp.csr_matrix(Bi)
            Ad = sp.csr_matrix(Ad)
            Bd = sp.csr_matrix(Bd)

        return Ai, Bi, Ad, Bd


class GaussLobattoGrid(GridData):
    """
    A GridData object that provides the node information for a Gauss-Lobatto distribution.

    Parameters
    ----------
    num_segments : int
        The number of segments in the phase.
    nodes_per_seg : int or iterable
        The number of nodes in each segment. As an integer, it applies to each segment. If a sequence, its length
        must be equal to num_segments.
    segment_ends : Iterable[num_segments + 1] or None
        The segments nodes on some arbitrary interval.
        This will be normalized to the interval [-1, 1].
    compressed : bool
        If the transcription is compressed, then states and controls at shared
        nodes of adjacent segments are only specified once, and then broadcast
        to the appropriate indices.
    """
    def __init__(self, num_segments, nodes_per_seg, segment_ends=None, compressed=False):
        self.grid_type = 'lgl'
        super().__init__(num_segments=num_segments, transcription='gauss-lobatto',
                         transcription_order=np.asarray(nodes_per_seg, dtype=int),
                         segment_ends=segment_ends, compressed=compressed)


class ChebyshevGaussLobattoGrid(GridData):
    """
    A GridData object that provides the node information for a Gauss-Lobatto distribution.

    Parameters
    ----------
    num_segments : int
        The number of segments in the phase.
    nodes_per_seg : int or iterable
        The number of nodes in each segment. As an integer, it applies to each segment. If a sequence, its length
        must be equal to num_segments.
    segment_ends : Iterable[num_segments + 1] or None
        The segments nodes on some arbitrary interval.
        This will be normalized to the interval [-1, 1].
    compressed : bool
        If the transcription is compressed, then states and controls at shared
        nodes of adjacent segments are only specified once, and then broadcast
        to the appropriate indices.
    """
    def __init__(self, num_segments, nodes_per_seg, segment_ends=None, compressed=False):
        self.grid_type = 'cgl'
        super().__init__(num_segments=num_segments, transcription='chebyshev-gauss-lobatto',
                         transcription_order=np.asarray(nodes_per_seg, dtype=int),
                         segment_ends=segment_ends, compressed=compressed)


class BirkhoffGrid(GridData):
    """
    A GridData object that provides the node information for the Birkhoff transcription.

    Parameters
    ----------
    num_nodes : int
        The number of nodes in the grid.
    grid_type : str
        The type of Gaussian grid used for the transcription. May be 'lgl' or 'cgl'.
    """
    def __init__(self, num_nodes, grid_type='cgl'):
        self.grid_type = grid_type
        super().__init__(num_segments=1, transcription='birkhoff',
                         transcription_order=num_nodes)


class RadauGrid(GridData):
    """
    A GridData object that provides the node information for a Radau distribution.

    Parameters
    ----------
    num_segments : int
        The number of segments in the phase.
    nodes_per_seg : int or iterable
        The number of nodes in each segment. As an integer, it applies to each segment. If a sequence, its length
        must be equal to num_segments.
    segment_ends : Iterable[num_segments + 1] or None
        The segments nodes on some arbitrary interval.
        This will be normalized to the interval [-1, 1].
    compressed : bool
        If the transcription is compressed, then states and controls at shared
        nodes of adjacent segments are only specified once, and then broadcast
        to the appropriate indices.
    """
    def __init__(self, num_segments, nodes_per_seg, segment_ends=None, compressed=False):
        self.grid_type = 'lgr'
        super().__init__(num_segments=num_segments, transcription='radau-ps',
                         transcription_order=np.asarray(nodes_per_seg, dtype=int) - 1,
                         segment_ends=segment_ends, compressed=compressed)


class UniformGrid(GridData):
    """
    A GridData object that provides the node information for a uniform distribution.

    Parameters
    ----------
    num_segments : int
        The number of segments in the phase.
    nodes_per_seg : int or iterable
        The number of nodes in each segment. As an integer, it applies to each segment. If a sequence, its length
        must be equal to num_segments.
    segment_ends : Iterable[num_segments + 1] or None
        The segments nodes on some arbitrary interval.
        This will be normalized to the interval [-1, 1].
    compressed : bool
        If the transcription is compressed, then states and controls at shared
        nodes of adjacent segments are only specified once, and then broadcast
        to the appropriate indices.
    """
    def __init__(self, num_segments, nodes_per_seg, segment_ends=None, compressed=False):
        self.grid_type = 'uniform'
        super().__init__(num_segments=num_segments, transcription='uniform',
                         transcription_order=np.asarray(nodes_per_seg) - 1,
                         segment_ends=segment_ends, compressed=compressed)



import os
import warnings
import pathlib

import numpy as np

import openmdao.api as om
from .._options import options as dymos_options


def _get_phases_node_in_problem_metadata(node, path=""):
    """
    Find the phases node in the Problem metadata hierarchy.

    There is one node in the hierarchy that has the name 'phases'. Finding this
    node will be used to find information about all the phases in the model.
    This is a recursive function.

    Parameters
    ----------
    node : list
        Node in Problem metadata hierarchy.
    path : str
        The dotted string path name to the node in the Problem hierarchy. Used
        recursively to build up the path to a node.

    Returns
    -------
    tuple of a list and a string
        Returns the node and path name to the node, if found. Otherwise, returns (None, None).
    """
    for item in node:
        if item['name'] == 'phases':
            return item, f'{path}'
        else:
            if 'children' in item:
                children = item['children']
                name = item['name']
                if path:
                    new_path = f'{path}:{name}'
                else:
                    new_path = name
                phases_node, phase_node_path = _get_phases_node_in_problem_metadata(children,
                                                                                    new_path)
                if phases_node:
                    return phases_node, phase_node_path
    return None, None


def _mpl_timeseries_plots(time_units, var_units, phase_names, phases_node_path,
                          last_solution_case, last_simulation_case, plot_dir_path,
                          dpi, include_parameters):
    import matplotlib
    import matplotlib.pyplot as plt
    import matplotlib.lines as mlines
    import matplotlib.patches as mpatches

    # get ready to plot
    backend_save = plt.get_backend()
    plt.switch_backend('Agg')
    # use a colormap with 20 values
    cm = matplotlib.colormaps['tab20']
    plotfiles = []

    for var_name, var_unit in var_units.items():
        if (not include_parameters) and ("parameters:" in var_name):
            continue

        # start a new plot
        fig, ax = plt.subplots()

        # Get the labels
        time_label = f'time ({time_units[var_name]})'
        var_label = f'{var_name} ({var_unit})'
        title = f'timeseries.{var_name}'

        # add labels, title, and legend
        ax.set_xlabel(time_label)
        ax.set_ylabel(var_label)
        fig.suptitle(title)

        # Plot each phase
        for iphase, phase_name in enumerate(phase_names):
            if phases_node_path:
                var_name_full = f'{phases_node_path}.{phase_name}.timeseries.{var_name}'
                time_name = f'{phases_node_path}.{phase_name}.timeseries.time'
            else:
                var_name_full = f'{phase_name}.timeseries.{var_name}'
                time_name = f'{phase_name}.timeseries.time'

            # Get values
            if var_name_full not in last_solution_case.outputs:
                continue

            var_val = last_solution_case.outputs[var_name_full]
            time_val = last_solution_case.outputs[time_name]

            # Plot the data
            color = cm.colors[iphase % 20]

            ax.plot(time_val, var_val, marker='o', linestyle='None', label='solution', color=color)

            # get simulation values, if plotting simulation
            if last_simulation_case:
                # if the phases_node_path is empty, need to pre-pend names with "sim_traj."
                #   as that is pre-pended in Trajectory.simulate code
                sim_prefix = "" if phases_node_path else "sim_traj."
                var_val_simulate = last_simulation_case.outputs[sim_prefix + var_name_full]
                time_val_simulate = last_simulation_case.outputs[sim_prefix + time_name]
                ax.plot(time_val_simulate, var_val_simulate, linestyle='--', label='simulation',
                        color=color)

        # Create two legends
        #   Solution/Simulation legend
        solution_line = mlines.Line2D([], [], color='black', marker='o', linestyle='None',
                                      label='Solution')
        if last_simulation_case:
            simulation_line = mlines.Line2D([], [], color='black', linestyle='--',
                                            label='Simulation')
            sol_sim_legend = plt.legend(handles=[solution_line, simulation_line],
                                        loc='upper left', bbox_to_anchor=(-0.3, -0.12), shadow=True)
        else:
            sol_sim_legend = plt.legend(handles=[solution_line],
                                        loc='upper left', bbox_to_anchor=(-0.3, -0.12),
                                        shadow=True)
        plt.gca().add_artist(sol_sim_legend)

        #   Phases legend
        handles = []
        for iphase, phase_name in enumerate(phase_names):
            patch = mpatches.Patch(color=cm.colors[iphase], label=phase_name)
            handles.append(patch)
        plt.legend(handles=handles, loc='upper right', ncol=len(phase_names), shadow=True,
                   bbox_to_anchor=(1.15, -0.12), title='Phases')

        plt.subplots_adjust(bottom=0.23, top=0.9, left=0.2)

        # save to file
        plot_file_path = pathlib.Path(plot_dir_path).joinpath(f'{var_name.replace(":", "_")}.png')
        plt.savefig(plot_file_path, dpi=dpi)
        plt.close(fig)
        plotfiles.append(plot_file_path)

    plt.switch_backend(backend_save)

    return plotfiles


def _bokeh_timeseries_plots(time_units, var_units, phase_names, phases_node_path,
                            last_solution_case, last_simulation_case, plot_dir_path, num_cols=2,
                            bg_fill_color='#282828', grid_line_color='#666666', open_browser=False):
    from bokeh.io import output_notebook, output_file, save, show
    from bokeh.layouts import gridplot, column
    from bokeh.models import Legend
    from bokeh.plotting import figure
    import bokeh.palettes as bp

    if dymos_options['notebook_mode']:
        output_notebook()
    else:
        output_file(os.path.join(plot_dir_path, 'plots.html'))

    # Prune the edges from the color map
    cmap = bp.turbo(len(phase_names) + 2)[1:-1]
    figures = []
    colors = {}
    sol_plots = {}
    sim_plots = {}

    # Get the minimum and maximum times in any phase, so when we plot a variable that only exists
    # in a few phases, it is plotted against the entire time range.
    min_time = 1.0E21
    max_time = -1.0E21

    for iphase, phase_name in enumerate(phase_names):
        if phases_node_path:
            time_name = f'{phases_node_path}.{phase_name}.timeseries.time'
        else:
            time_name = f'{phase_name}.timeseries.time'
        min_time = min(min_time, np.min(last_solution_case.outputs[time_name]))
        max_time = max(max_time, np.max(last_solution_case.outputs[time_name]))
        colors[phase_name] = cmap[iphase]

    for var_name, var_unit in var_units.items():
        # Get the labels
        time_label = f'time ({time_units[var_name]})'
        var_label = f'{var_name} ({var_unit})'
        title = f'timeseries.{var_name}'

        # add labels, title, and legend
        padding = 0.05 * (max_time - min_time)
        fig = figure(title=title, background_fill_color=bg_fill_color,
                     x_range=(min_time - padding, max_time + padding),
                     width=180, height=180)
        fig.xaxis.axis_label = time_label
        fig.yaxis.axis_label = var_label
        fig.xgrid.grid_line_color = grid_line_color
        fig.ygrid.grid_line_color = grid_line_color

        # Plot each phase
        for iphase, phase_name in enumerate(phase_names):
            sol_color = cmap[iphase]
            sim_color = cmap[iphase]

            if phases_node_path:
                var_name_full = f'{phases_node_path}.{phase_name}.timeseries.{var_name}'
                time_name = f'{phases_node_path}.{phase_name}.timeseries.time'
            else:
                var_name_full = f'{phase_name}.timeseries.{var_name}'
                time_name = f'{phase_name}.timeseries.time'

            # Get values
            if var_name_full not in last_solution_case.outputs:
                continue

            var_val = last_solution_case.outputs[var_name_full]
            time_val = last_solution_case.outputs[time_name]

            for idxs, i in np.ndenumerate(np.zeros(var_val.shape[1:])):
                var_val_i = var_val[:, idxs].ravel()
                sol_plots[phase_name] = fig.circle(time_val.ravel(), var_val_i, size=5,
                                                   color=sol_color, name='sol:' + phase_name)

            # get simulation values, if plotting simulation
            if last_simulation_case:
                # if the phases_node_path is empty, need to pre-pend names with "sim_traj."
                #   as that is pre-pended in Trajectory.simulate code
                sim_prefix = '' if phases_node_path else 'sim_traj.'
                var_val_simulate = last_simulation_case.outputs[sim_prefix + var_name_full]
                time_val_simulate = last_simulation_case.outputs[sim_prefix + time_name]
                for idxs, i in np.ndenumerate(np.zeros(var_val_simulate.shape[1:])):
                    var_val_i = var_val_simulate[:, idxs].ravel()
                    sim_plots[phase_name] = fig.line(time_val_simulate.ravel(), var_val_i,
                                                     line_dash='solid', line_width=0.5, color=sim_color,
                                                     name='sim:' + phase_name)
        figures.append(fig)

    # Implement a single legend for all figures using the example here:
    # https://stackoverflow.com/a/56825812/754536

    # ## Use a dummy figure for the LEGEND
    dum_fig = figure(outline_line_alpha=0, toolbar_location=None,
                     background_fill_color=bg_fill_color, width=250, max_width=250)

    # set the components of the figure invisible
    for fig_component in [dum_fig.grid, dum_fig.ygrid, dum_fig.xaxis, dum_fig.yaxis]:
        fig_component.visible = False

    # The glyphs referred by the legend need to be present in the figure that holds the legend,
    # so we must add them to the figure renderers.
    sol_legend_items = [(phase_name + ' solution', [dum_fig.circle([0], [0],
                                                                   size=5,
                                                                   color=colors[phase_name],
                                                                   tags=['sol:' + phase_name])]) for phase_name in phase_names]
    sim_legend_items = [(phase_name + ' simulation', [dum_fig.line([0], [0],
                                                                   line_dash='solid',
                                                                   line_width=0.5,
                                                                   color=colors[phase_name],
                                                                   tags=['sim:' + phase_name])]) for phase_name in phase_names]
    legend_items = [j for i in zip(sol_legend_items, sim_legend_items) for j in i]

    # # set the figure range outside of the range of all glyphs
    dum_fig.x_range.end = 1005
    dum_fig.x_range.start = 1000

    legend = Legend(click_policy='hide', location='top_left', border_line_alpha=0, items=legend_items,
                    background_fill_alpha=0.0, label_text_color='white', label_width=120, spacing=10)

    dum_fig.add_layout(legend, place='center')

    gd = gridplot(figures, ncols=num_cols, sizing_mode='scale_both')

    plots = gridplot([[gd, column(dum_fig, sizing_mode='stretch_height')]],
                     toolbar_location=None,
                     sizing_mode='scale_both')

    if dymos_options['notebook_mode'] or open_browser:
        show(plots)
    else:
        save(plots)


try:
    from openmdao.utils.file_utils import image2html
except ImportError:
    # FIXME: this function is defined in newer versions of openmdao, so remove this
    # at some point later when we don't care about testing vs. older (< 3.19) openmdao versions.
    def image2html(imagefile, title='', alt=''):
        """
        Wrap the given image for display as an html file.

        Returns an html syntax string that can be written to a file.

        Parameters
        ----------
        imagefile : str
            Name of image file to be displayed.
        title : str
            The page title.
        alt : str
            Set the alt text for the image.

        Returns
        -------
        str
            Content string to create an html file.
        """
        return """
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <style>
            h2 {text-align: center;}
            .center {
                display: block;
                margin-left: auto;
                margin-right: auto;
                width: 80%;
            }
        </style>
    </head>
    <body>
    <h2>""" + title + "</h2>" + f"""
    <img src="{imagefile}" alt="{alt}" class="center"></img>

    </body>
    </html>
    """


def timeseries_plots(solution_recorder_filename, simulation_record_file=None, plot_dir="plots",
                     problem=None, dpi=150, make_html=True, include_parameters=True):
    """
    Create plots of the timeseries.

    Given timeseries data from case recorder files, make separate plots of each variable
    and store the plot files in the directory indicated by the variable plot_dir

    Parameters
    ----------
    solution_recorder_filename : str
        The path to the case recorder file containing solution data.
    simulation_record_file : str or None (default:None)
        The path to the case recorder file containing simulation data. If not None,
        this implies that the data from it should be plotted.
    plot_dir : str
        The path to the directory to which the plot files will be written.
    problem : Problem or None
        If not None, this is the owning Problem, and the plot_dir will be relative to the reports
        directory for this Problem.
    dpi : float
        The dpi (pixels per inch) for the matplotlib images to be saved. A larger dpi number
        results in higher resolution images.
    make_html : bool
        If true, make .html files that wrap the generated images.
    include_parameters : bool
        If true, include parameters in the timeseries plots. It can be helpful to set this to false
        for models with only static parameters that are uninteresting to plot.
    """
    # get ready to generate plot files
    if problem is not None:
        plot_dir = problem.get_reports_dir() / plot_dir
    pathlib.Path(plot_dir).mkdir(parents=True, exist_ok=True)

    cr = om.CaseReader(solution_recorder_filename)

    # get outputs from the solution
    solution_cases = cr.list_cases('problem', out_stream=None)
    last_solution_case = cr.get_case(solution_cases[-1])

    # If plotting simulation results, get the values for those variables
    if simulation_record_file:
        cr_simulate = om.CaseReader(simulation_record_file)
        system_simulation_cases = cr_simulate.list_cases('problem', out_stream=None)
        last_simulation_case = cr_simulate.get_case(system_simulation_cases[-1])
    else:
        last_simulation_case = None

    # we will use the problem metadata to get information about the phases and units
    root_children = cr.problem_metadata['tree']['children']

    # We have to key off the phases node. It will tell us
    #  what the prefix is for the variable names before the phases part of the variables
    #  and also let us know what the phase names are
    phases_node, phases_node_path = _get_phases_node_in_problem_metadata(root_children)

    # get phase names
    phase_names = [phase['name'] for phase in phases_node['children']]

    # Get the units and var names along the way
    var_units = {}
    time_units = {}
    for phase_node in phases_node['children']:  # Hopefully all phases have the same vars
        for phase_node_child in phase_node['children']:  # find the timeseries node
            if phase_node_child['name'] == 'timeseries':
                timeseries_node = phase_node_child
                break
        # get the time units first so they can be associated with all the variables
        for timeseries_node_child in timeseries_node['children']:
            if timeseries_node_child['name'] == 'time':
                units_for_time = timeseries_node_child['units']
                break
        for timeseries_node_child in timeseries_node['children']:
            # plot everything in the timeseries except input_values. Also not time since that
            #   is the independent variable in these plot
            if not timeseries_node_child['name'].startswith("input_values:")\
                    and not timeseries_node_child['name'] == 'time':
                varname = timeseries_node_child['name']
                var_units[varname] = timeseries_node_child['units']
                time_units[varname] = units_for_time

    # Check to see if there is anything to plot
    if len(var_units) == 0:
        warnings.warn('There are no timeseries variables to plot', RuntimeWarning)
        return

    if dymos_options['plots'] == 'bokeh':
        _bokeh_timeseries_plots(time_units, var_units, phase_names, phases_node_path,
                                last_solution_case, last_simulation_case, plot_dir)
    elif dymos_options['plots'] == 'matplotlib':
        fnames = _mpl_timeseries_plots(time_units, var_units, phase_names, phases_node_path,
                                       last_solution_case, last_simulation_case, plot_dir,
                                       dpi, include_parameters)
        if (problem is not None) and make_html:
            for name in fnames:
                # create html files that wrap the image files
                fpath = pathlib.Path(name).resolve()
                htmlpath = str(fpath.parent.joinpath(fpath.stem + '.html'))
                with open(htmlpath, 'w', encoding='utf-8') as f:
                    f.write(image2html(fpath.name))
    else:
        raise ValueError(f'Unknown plotting option: {dymos_options["plots"]}')



import numpy as np
import openmdao.api as om


class BrachistochroneVectorStatesODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=9.80665 * np.ones(nn), desc='grav. acceleration', units='m/s/s')

        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')

        self.add_output('pos_dot', val=np.zeros((nn, 2)), desc='velocity components', units='m/s',
                        tags=['dymos.state_rate_source:pos'])

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2',
                        tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])

        self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                        units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        rs = np.arange(2*nn, dtype=int)
        cs = np.repeat(np.arange(nn, dtype=int), 2)
        self.declare_partials(of='pos_dot', wrt='v', rows=rs, cols=cs)
        self.declare_partials(of='pos_dot', wrt='theta', rows=rs, cols=cs)

        self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='check', wrt='theta', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['pos_dot'][:, 0] = v * sin_theta
        outputs['pos_dot'][:, 1] = -v * cos_theta
        outputs['check'] = v / sin_theta

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        jacobian['vdot', 'g'] = cos_theta
        jacobian['vdot', 'theta'] = -g * sin_theta

        jacobian['pos_dot', 'v'][0::2] = sin_theta
        jacobian['pos_dot', 'v'][1::2] = -cos_theta

        jacobian['pos_dot', 'theta'][0::2] = v * cos_theta
        jacobian['pos_dot', 'theta'][1::2] = v * sin_theta

        jacobian['check', 'v'] = 1 / sin_theta
        jacobian['check', 'theta'] = -v * cos_theta / sin_theta**2



import os
import unittest

import numpy as np

import openmdao.api as om
import dymos as dm

from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

OPT, OPTIMIZER = set_pyoptsparse_opt('SNOPT', fallback=True)


@use_tempdirs
class TestDuplicateConstraints(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['phase0_sim.db', 'brachistochrone_sim.db']:
            if os.path.exists(filename):
                os.remove(filename)

    def run_asserts(self, p, tol=0.01):

        t_initial = p.get_val('traj0.phase0.timeseries.time')[0]
        tf = p.get_val('traj0.phase0.timeseries.time')[-1]

        x0 = p.get_val('traj0.phase0.timeseries.x')[0]
        xf = p.get_val('traj0.phase0.timeseries.x')[-1]

        y0 = p.get_val('traj0.phase0.timeseries.y')[0]
        yf = p.get_val('traj0.phase0.timeseries.y')[-1]

        v0 = p.get_val('traj0.phase0.timeseries.v')[0]
        vf = p.get_val('traj0.phase0.timeseries.v')[-1]

        g = p.get_val('traj0.phase0.timeseries.g')[0]

        thetaf = p.get_val('traj0.phase0.timeseries.theta', units='deg')[-1]

        assert_near_equal(t_initial, 0.0, tolerance=0.01)
        assert_near_equal(x0, 0.0, tolerance=0.01)
        assert_near_equal(y0, 10.0, tolerance=0.01)
        assert_near_equal(v0, 0.0, tolerance=0.01)

        assert_near_equal(tf, 1.8016, tolerance=0.01)
        assert_near_equal(xf, 10.0, tolerance=0.01)
        assert_near_equal(yf, 5.0, tolerance=0.01)
        assert_near_equal(vf, 9.902, tolerance=0.01)
        assert_near_equal(g, 9.80665, tolerance=0.01)

        assert_near_equal(thetaf, 100.12, tolerance=0.01)

    def test_duplicate_initial_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0)

        # Now mistakenly add a boundary constraint at the same loc, variable, and indices.
        with self.assertRaises(ValueError) as e:
            phase.add_boundary_constraint('x', loc='initial', equals=10)

        expected = 'Cannot add new initial boundary constraint for variable `x` and indices None. One already exists.'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_final_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=0)

        # Now mistakenly add a boundary constraint at the same loc, variable, and indices.
        with self.assertRaises(ValueError) as e:
            phase.add_boundary_constraint('x', loc='final', equals=10)

        expected = 'Cannot add new final boundary constraint for variable `x` and indices None. One already exists.'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_path_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_path_constraint('theta', lower=0)

        # Now mistakenly add a boundary constraint at the same loc, variable, and indices.
        with self.assertRaises(ValueError) as e:
            phase.add_path_constraint('theta', upper=100)

        expected = 'Cannot add new path constraint for variable `theta` and indices None. One already exists.'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_initial_constraint_equivalent_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0)
        phase.add_boundary_constraint('y', loc='initial', equals=10)
        phase.add_boundary_constraint('v', loc='initial', equals=0)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        # Now mistakenly add a boundary constraint at the same loc, variable, and different but equivalent indices.
        phase.add_boundary_constraint('x', loc='initial', equals=10, indices=[0])

        phase.add_objective('time', loc='final')

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `x` are ' \
                   'used in multiple initial boundary constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_final_constraint_equivalent_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0)
        phase.add_boundary_constraint('y', loc='initial', equals=10)
        phase.add_boundary_constraint('v', loc='initial', equals=0)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        # Now mistakenly add a boundary constraint at the same loc, variable, and different but equivalent indices.
        phase.add_boundary_constraint('x', loc='final', equals=10, indices=[0])

        phase.add_objective('time', loc='final')

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `x` are ' \
                   'used in multiple final boundary constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_path_constraint_equivalent_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0)
        phase.add_boundary_constraint('y', loc='initial', equals=10)
        phase.add_boundary_constraint('v', loc='initial', equals=0)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        # Now mistakenly add a path constraint at the same loc, variable, and different but equivalent indices
        phase.add_path_constraint('theta', upper=100)
        phase.add_path_constraint('theta', lower=0, indices=[0])

        phase.add_objective('time', loc='final')

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `theta` are ' \
                   'used in multiple path constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_initial_constraint_equivalent_negative_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0)

        # Now mistakenly add a boundary constraint at the same loc, variable, and different but equivalent negative indices.
        phase.add_boundary_constraint('x', loc='initial', equals=10, indices=[-1])

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `x` are ' \
                   'used in multiple initial boundary constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_final_constraint_equivalent_negative_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='final', equals=10)

        # Now mistakenly add a boundary constraint at the same loc, variable, and different but equivalent negative indices.
        phase.add_boundary_constraint('x', loc='final', equals=10, indices=[-1])

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `x` are ' \
                   'used in multiple final boundary constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_path_constraint_equivalent_negative_indices(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        # Now mistakenly add a path constraint at the same loc, variable, and different but equivalent negative indices
        phase.add_path_constraint('theta', upper=100)
        phase.add_path_constraint('theta', lower=0, indices=[-1])

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `theta` are ' \
                   'used in multiple path constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_initial_constraint_indies_as_ndarray(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2')

        phase.add_boundary_constraint('x', loc='initial', equals=0, indices=np.array([0], dtype=int))

        # Now mistakenly add a boundary constraint at the same loc, variable, and different but equivalent negative indices.
        phase.add_boundary_constraint('x', loc='initial', equals=10, indices=np.array([-1], dtype=int))

        with self.assertRaises(ValueError) as e:
            p.setup()

        expected = 'Duplicate constraint in phase traj0.phases.phase0. The following indices of `x` are ' \
                   'used in multiple initial boundary constraints:\n{0}'

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_parameter_constraint(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=False, fix_final=False)
        phase.add_state('y', fix_initial=False, fix_final=False)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=False, fix_final=False)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], units='m/s**2', opt=True)

        # Add a parameter as both a boundary and path constraint. This is illegal since the parameter value cannot
        # change during the phase.  A user must choose either a path constraint or a boundary constraint.
        phase.add_boundary_constraint('g', loc='initial', upper=9.80665)
        phase.add_path_constraint('g', lower=0)

        with self.assertRaises(RuntimeError) as e:
            p.setup()

        expected = ("In phase traj0.phases.phase0, parameter `g` is subject to multiple boundary or path constraints.\n"
                    "Parameters are single values that do not change in time, and may only be used in a single "
                    "boundary or path constraint.")

        self.assertEqual(str(e.exception), expected)

    def test_duplicate_path_constraint_different_constraint_name(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver(tol=1.0E-6)
        p.driver.declare_coloring(tol=1.0E-12)

        tx = dm.Radau(num_segments=10, order=3)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='rad', lower=1E-6, upper=np.pi)

        phase.add_parameter('g', units='m/s**2')

        # Constrain the ODE output "check", but with different names in different places.
        phase.add_boundary_constraint('check', loc='final', lower=-50, upper=50)
        phase.add_path_constraint('check', constraint_name='bounded_check', upper=100, lower=-100)
        phase.add_objective('time', loc='final')

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0.001, 9.9])
        phase.set_control_val('theta', [5, 90], units='deg')
        phase.set_parameter_val('g', 9.80665)

        p.run_model()

        v = p.get_val('traj0.phase0.timeseries.v')
        theta = p.get_val('traj0.phase0.timeseries.theta', units='rad')

        check_calc = v / np.sin(theta)
        check_1 = p.get_val('traj0.phase0.timeseries.check')
        check_2 = p.get_val('traj0.phase0.timeseries.bounded_check')

        assert_near_equal(check_calc, check_1)
        assert_near_equal(check_calc, check_2)
